@@ -10,6 +10,7 @@ from scipy.sparse.linalg.interface import LinearOperator
 from scipy._lib.decorator import decorator
 from .utils import make_system
 from scipy._lib._util import _aligned_zeros
+from scipy._lib._threadsafety import non_reentrant
 
 _type_conv = {'f':'s', 'd':'d', 'F':'c', 'D':'z'}
 
@@ -53,15 +54,6 @@ M : {sparse matrix, dense matrix, LinearOperator}
 callback : function
     User-supplied function to call after each iteration.  It is called
     as callback(xk), where xk is the current solution vector.
-xtype : {'f','d','F','D'}
-    This parameter is deprecated -- avoid using it.
-
-    The type of the result.  If None, then it will be determined from
-    A.dtype.char and b.  If A does not have a typecode method then it
-    will compute A.matvec(x0) to get a typecode.   To save the extra
-    computation when A does not have a typecode attribute use xtype=0
-    for the same type as b or use xtype='f','d','F',or 'D'.
-    This parameter has been superseded by LinearOperator.
 
 """
 
@@ -75,25 +67,13 @@ def set_docstring(header, Ainfo, footer=''):
     return combine
 
 
-@decorator
-def non_reentrant(func, *a, **kw):
-    d = func.__dict__
-    if d.get('__entered'):
-        raise RuntimeError("%s is not re-entrant" % func.__name__)
-    try:
-        d['__entered'] = True
-        return func(*a, **kw)
-    finally:
-        d['__entered'] = False
-
-
-@set_docstring('Use BIConjugate Gradient iteration to solve A x = b',
-               'The real or complex N-by-N matrix of the linear system\n'
+@set_docstring('Use BIConjugate Gradient iteration to solve ``Ax = b``.',
+               'The real or complex N-by-N matrix of the linear system.\n'
                'It is required that the linear operator can produce\n'
                '``Ax`` and ``A^T x``.')
-@non_reentrant
-def bicg(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None):
-    A,M,x,b,postprocess = make_system(A,M,x0,b,xtype)
+@non_reentrant()
+def bicg(A, b, x0=None, tol=1e-5, maxiter=None, M=None, callback=None):
+    A,M,x,b,postprocess = make_system(A, M, x0, b)
 
     n = len(b)
     if maxiter is None:
@@ -154,12 +134,12 @@ def bicg(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=Non
     return postprocess(x), info
 
 
-@set_docstring('Use BIConjugate Gradient STABilized iteration to solve A x = b',
-               'The real or complex N-by-N matrix of the linear system\n'
-               '``A`` must represent a hermitian, positive definite matrix')
-@non_reentrant
-def bicgstab(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None):
-    A,M,x,b,postprocess = make_system(A,M,x0,b,xtype)
+@set_docstring('Use BIConjugate Gradient STABilized iteration to solve '
+               '``Ax = b``.',
+               'The real or complex N-by-N matrix of the linear system.')
+@non_reentrant()
+def bicgstab(A, b, x0=None, tol=1e-5, maxiter=None, M=None, callback=None):
+    A, M, x, b, postprocess = make_system(A, M, x0, b)
 
     n = len(b)
     if maxiter is None:
@@ -215,12 +195,12 @@ def bicgstab(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback
     return postprocess(x), info
 
 
-@set_docstring('Use Conjugate Gradient iteration to solve A x = b',
-               'The real or complex N-by-N matrix of the linear system\n'
-               '``A`` must represent a hermitian, positive definite matrix')
-@non_reentrant
-def cg(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None):
-    A,M,x,b,postprocess = make_system(A,M,x0,b,xtype)
+@set_docstring('Use Conjugate Gradient iteration to solve ``Ax = b``.',
+               'The real or complex N-by-N matrix of the linear system.\n'
+               '``A`` must represent a hermitian, positive definite matrix.')
+@non_reentrant()
+def cg(A, b, x0=None, tol=1e-5, maxiter=None, M=None, callback=None):
+    A, M, x, b, postprocess = make_system(A, M, x0, b)
 
     n = len(b)
     if maxiter is None:
@@ -276,11 +256,11 @@ def cg(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None)
     return postprocess(x), info
 
 
-@set_docstring('Use Conjugate Gradient Squared iteration to solve A x = b',
-               'The real-valued N-by-N matrix of the linear system')
-@non_reentrant
-def cgs(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None):
-    A,M,x,b,postprocess = make_system(A,M,x0,b,xtype)
+@set_docstring('Use Conjugate Gradient Squared iteration to solve ``Ax = b``.',
+               'The real-valued N-by-N matrix of the linear system.')
+@non_reentrant()
+def cgs(A, b, x0=None, tol=1e-5, maxiter=None, M=None, callback=None):
+    A, M, x, b, postprocess = make_system(A, M, x0, b)
 
     n = len(b)
     if maxiter is None:
@@ -336,10 +316,10 @@ def cgs(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M=None, callback=None
     return postprocess(x), info
 
 
-@non_reentrant
-def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, xtype=None, M=None, callback=None, restrt=None):
+@non_reentrant()
+def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, M=None, callback=None, restrt=None):
     """
-    Use Generalized Minimal RESidual iteration to solve A x = b.
+    Use Generalized Minimal RESidual iteration to solve ``Ax = b``.
 
     Parameters
     ----------
@@ -373,15 +353,6 @@ def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, xtype=None, M=Non
         Maximum number of iterations (restart cycles).  Iteration will stop
         after maxiter steps even if the specified tolerance has not been
         achieved.
-    xtype : {'f','d','F','D'}
-        This parameter is DEPRECATED --- avoid using it.
-
-        The type of the result.  If None, then it will be determined from
-        A.dtype.char and b.  If A does not have a typecode method then it
-        will compute A.matvec(x0) to get a typecode.   To save the extra
-        computation when A does not have a typecode attribute use xtype=0
-        for the same type as b or use xtype='f','d','F',or 'D'.
-        This parameter has been superseded by LinearOperator.
     M : {sparse matrix, dense matrix, LinearOperator}
         Inverse of the preconditioner of A.  M should approximate the
         inverse of A and be easy to solve for (see Notes).  Effective
@@ -410,6 +381,17 @@ def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, xtype=None, M=Non
       M_x = lambda x: spla.spsolve(P, x)
       M = spla.LinearOperator((n, n), M_x)
 
+    Examples
+    --------
+    >>> from scipy.sparse import csc_matrix
+    >>> from scipy.sparse.linalg import gmres
+    >>> A = csc_matrix([[3, 2, 0], [1, -1, 0], [0, 5, 1]], dtype=float)
+    >>> b = np.array([2, 4, -1], dtype=float)
+    >>> x, exitCode = gmres(A, b)
+    >>> print(exitCode)            # 0 indicates successful convergence
+    0
+    >>> np.allclose(A.dot(x), b)
+    True
     """
 
     # Change 'restrt' keyword to 'restart'
@@ -419,7 +401,7 @@ def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, xtype=None, M=Non
         raise ValueError("Cannot specify both restart and restrt keywords. "
                          "Preferably use 'restart' only.")
 
-    A,M,x,b,postprocess = make_system(A,M,x0,b,xtype)
+    A, M, x, b,postprocess = make_system(A, M, x0, b)
 
     n = len(b)
     if maxiter is None:
@@ -500,9 +482,9 @@ def gmres(A, b, x0=None, tol=1e-5, restart=None, maxiter=None, xtype=None, M=Non
     return postprocess(x), info
 
 
-@non_reentrant
-def qmr(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M1=None, M2=None, callback=None):
-    """Use Quasi-Minimal Residual iteration to solve A x = b
+@non_reentrant()
+def qmr(A, b, x0=None, tol=1e-5, maxiter=None, M1=None, M2=None, callback=None):
+    """Use Quasi-Minimal Residual iteration to solve ``Ax = b``.
 
     Parameters
     ----------
@@ -542,23 +524,25 @@ def qmr(A, b, x0=None, tol=1e-5, maxiter=None, xtype=None, M1=None, M2=None, cal
     callback : function
         User-supplied function to call after each iteration.  It is called
         as callback(xk), where xk is the current solution vector.
-    xtype : {'f','d','F','D'}
-        This parameter is DEPRECATED -- avoid using it.
-
-        The type of the result.  If None, then it will be determined from
-        A.dtype.char and b.  If A does not have a typecode method then it
-        will compute A.matvec(x0) to get a typecode.   To save the extra
-        computation when A does not have a typecode attribute use xtype=0
-        for the same type as b or use xtype='f','d','F',or 'D'.
-        This parameter has been superseded by LinearOperator.
 
     See Also
     --------
     LinearOperator
 
+    Examples
+    --------
+    >>> from scipy.sparse import csc_matrix
+    >>> from scipy.sparse.linalg import qmr
+    >>> A = csc_matrix([[3, 2, 0], [1, -1, 0], [0, 5, 1]], dtype=float)
+    >>> b = np.array([2, 4, -1], dtype=float)
+    >>> x, exitCode = qmr(A, b)
+    >>> print(exitCode)            # 0 indicates successful convergence
+    0
+    >>> np.allclose(A.dot(x), b)
+    True
     """
     A_ = A
-    A,M,x,b,postprocess = make_system(A,None,x0,b,xtype)
+    A, M, x, b, postprocess = make_system(A, None, x0, b)
 
     if M1 is None and M2 is None:
         if hasattr(A_,'psolve'):
